import torch
from torch.utils.tensorboard.writer import SummaryWriter
from pathlib import Path
import os
import json
from typing import (
    List,
    NamedTuple,
    Optional,
    TYPE_CHECKING,
)
from shutil import rmtree
from xpmir.context import InitializationHook, Hook
from xpmir.utils.utils import easylog
from xpmir.learning.devices import DeviceInformation, ComputationContext
from xpmir.learning.metrics import Metric, Metrics
from experimaestro.utils import cleanupdir
from contextlib import contextmanager

if TYPE_CHECKING:
    from xpmir.learning.optim import ScheduledOptimizer, Module
    from xpmir.learning.trainers import Trainer

logger = easylog()


class Loss(NamedTuple):
    """A loss"""

    name: str
    value: torch.Tensor
    weight: float


class TrainState:
    """Represents a training state for serialization"""

    MODEL_PATH = "model.pth"

    epoch: int
    """The epoch"""

    steps: int
    """The number of steps (each epoch is composed of sptes)"""

    def __init__(
        self,
        model: "Module",
        trainer: "Trainer",
        optimizer: "ScheduledOptimizer",
        epoch=0,
        steps=0,
    ):
        # Initialize the state
        self.model = model
        self.trainer = trainer
        self.optimizer = optimizer

        self.epoch = epoch
        self.steps = steps

        # Was it loaded from disk?
        self.cached = False

        # Was it saved?
        self.path = None

    def copy(self):
        return TrainState(self.model, self.trainer, self.optimizer, **self.state_dict())

    def state_dict(self):
        return {
            "epoch": self.epoch,
            "steps": self.steps,
        }

    @property
    def step(self):
        """Returns the step for logging (number of steps)"""
        return self.steps

    def load_state_dict(self, state):
        self.epoch = state.get("epoch", 0)
        self.steps = state.get("steps", 0)

    def save(self, path):
        """Save the state"""
        cleanupdir(path)

        with (path / "info.json").open("wt") as fp:
            json.dump(self.state_dict(), fp)

        torch.save(self.model.state_dict(), path / TrainState.MODEL_PATH)
        torch.save(self.trainer.state_dict(), path / "trainer.pth")
        torch.save(self.optimizer.state_dict(), path / "optimizer.pth")

        self.path = path

    def load(self, path, onlyinfo=False):
        """Loads the state from disk"""
        if not onlyinfo:
            self.model.load_state_dict(torch.load(path / TrainState.MODEL_PATH))
            self.trainer.load_state_dict(torch.load(path / "trainer.pth"))
            self.optimizer.load_state_dict(torch.load(path / "optimizer.pth"))

        with (path / "info.json").open("rt") as fp:
            self.load_state_dict(json.load(fp))
        self.path = path
        self.cached = True

    def copy_model(self, path: Path):
        assert self.path is not None
        for name in [TrainState.MODEL_PATH, "info.json"]:
            os.link(self.path / name, path / name)


class TrainingHook(Hook):
    """Base class for all training hooks"""

    pass


class ValidationHook(Hook):
    """Base class for all the validation hooks"""

    def after(self, state: "TrainerContext"):
        """Called after a validation step"""

    def before(self, state: "TrainerContext"):
        """Called before a validation step"""


class StepTrainingHook(TrainingHook):
    """Base class for hooks called at each step (before/after)"""

    def after(self, state: "TrainerContext"):
        """Called after a training step"""

    def before(self, state: "TrainerContext"):
        """Called before a training step"""


class InitializationTrainingHook(TrainingHook, InitializationHook):
    """Base class for hooks called at initialization"""

    def after(self, state: "TrainerContext"):
        pass

    def before(self, state: "TrainerContext"):
        pass


class TrainerContext(ComputationContext):
    """Contains all the information about the training context
    for a spefic

    This object is used when training to provide models and losses'
    with extra information - as well as the possibility to add
    regularization losses
    """

    metrics: Optional[Metrics]
    """Metrics to be reported"""

    _losses: Optional[List[Loss]]
    """Regularization losses to be added to the main loss"""

    _scope: List[str]
    """Scope for metric names"""

    PREFIX = "epoch-"

    def __init__(
        self,
        device_information: DeviceInformation,
        logpath: Path,
        path: Path,
        max_epoch: int,
        steps_per_epoch: int,
        trainer,
        model: "Module",
        optimizer: "ScheduledOptimizer",
    ):
        super().__init__()
        self.device_information = device_information
        self.path = path
        self.logpath = logpath
        self.max_epoch = max_epoch
        self.steps_per_epoch = steps_per_epoch
        self._writer = None
        self._scope = []
        self._losses = None

        self.state = TrainState(model, trainer, optimizer)

    @property
    def writer(self):
        """Returns a tensorboard writer

        by default, purges the entries beside the current epoch
        """
        if self._writer is None:
            self._writer = SummaryWriter(self.logpath, purge_step=self.state.step)
        return self._writer

    @property
    def epoch(self):
        return self.state.epoch

    @property
    def steps(self):
        return self.state.steps

    def nextepoch(self):
        self.oldstate = self.state
        self.state = self.oldstate.copy()
        self.state.epoch += 1

    def nextbatch(self):
        self.state.steps += 1

    def load_bestcheckpoint(self, max_epoch: int):
        """Try to find the best checkpoint to load (the highest lower than
        the epoch target)"""
        # Find all the potential epochs
        epochs = []
        for f in self.path.glob(f"{TrainerContext.PREFIX}*"):
            epoch = int(f.name[len(TrainerContext.PREFIX) :])
            if epoch <= max_epoch:
                epochs.append(epoch)
        epochs.sort(reverse=True)

        # Try to load the first one
        for epoch in epochs:
            logger.info("Loading from checkpoint at epoch %d", epoch)
            path = self.path / f"{TrainerContext.PREFIX}{epoch:08d}"

            try:
                self.state.load(path)
                return True
            except NotImplementedError:
                logger.error("Not removing checkpoint")
                raise
            except Exception:
                rmtree(path)
                logger.exception("Cannot load from epoch %d", epoch)

        return False

    def save_checkpoint(self):
        # Serialize
        path = self.path / f"{TrainerContext.PREFIX}{self.epoch:08d}"
        if self.state.path is not None:
            # No need to save twice
            return

        # Save
        self.state.save(path)

        # Cleanup if needed
        if self.oldstate and self.oldstate.path:
            try:
                rmtree(self.oldstate.path)
            except OSError as e:
                # We continue the learning process in those cases
                logger.error("OS Error while trying to remove directory %s", e)

            self.oldstate = None

    def copy(self, path: Path):
        """Copy the state into another folder"""
        if self.state.path is None:
            self.save_checkpoint()

        trainpath = self.state.path
        assert trainpath is not None

        if path:
            cleanupdir(path)
            self.state.copy_model(path)

    def add_loss(self, loss: Loss):
        assert (
            self._losses is not None
        ), "This should be called in the context where loss is computed"
        self._losses.append(loss)

    @contextmanager
    def losses(self):
        previous = self._losses
        try:
            self._losses = []
            yield self._losses
        finally:
            self._losses = previous

    @contextmanager
    def step(self, metrics):
        try:
            self.state.optimizer.zero_grad()
            self.metrics = Metrics()
            yield self.metrics
            self.state.optimizer.optimizer_step(self)
            self.state.optimizer.scheduler_step(self)
            metrics.merge(self.metrics)
        finally:
            self.metrics = None

    def add_metric(self, metric: Metric):
        assert self.metrics is not None, "Not within an optimization step"
        if self._scope:
            metric.key = "/".join(s for s in self._scope if s) + "/" + metric.key
        self.metrics.add(metric)

    @contextmanager
    def scope(self, name: str):
        try:
            self._scope.append(name)
            yield self
        finally:
            self._scope.pop()



from datamaestro import prepare_dataset

queries = prepare_dataset("irds.aquaint.trec-robust-2005.queries")
for query in queries.iter():
    print(query.qid, query.text, query.metadata)  # (qid, text, metadata) tuple



import sys
from typing import Iterator
import torch
import torch.nn.functional as F
from experimaestro import Config, Param, initializer
from xpmir.letor.samplers import BatchwiseSampler, BatchwiseRecords
from xpmir.learning.context import Loss, TrainerContext
from xpmir.rankers import ScorerOutputType
from xpmir.letor.trainers import LossTrainer
import numpy as np


class BatchwiseLoss(Config):
    NAME = "?"

    weight: Param[float] = 1.0
    """The weight of this loss"""

    def initialize(self, context: TrainerContext):
        pass

    def process(
        self, scores: torch.Tensor, relevances: torch.Tensor, context: TrainerContext
    ):
        value = self.compute(scores, relevances, context)
        context.add_loss(Loss(f"batch-{self.NAME}", value, self.weight))

    def compute(
        self, scores: torch.Tensor, relevances: torch.Tensor, context: TrainerContext
    ) -> torch.Tensor:
        """
        Compute the loss

        Arguments:

        - scores: A (queries x documents) tensor
        - relevances: A (queries x documents) tensor
        """
        raise NotImplementedError()


class CrossEntropyLoss(BatchwiseLoss):
    NAME = "bce"

    def compute(self, scores, relevances, context):
        return F.binary_cross_entropy(scores, relevances, reduction="mean")


class SoftmaxCrossEntropy(BatchwiseLoss):
    NAME = "ce"

    """Computes the probability of relevant documents for a given query"""

    def initialize(self, context: TrainerContext):
        super().initialize(context)
        self.mode = context.state.model.outputType
        self.normalize = {
            ScorerOutputType.REAL: lambda x: F.log_softmax(x, -1),
            ScorerOutputType.LOG_PROBABILITY: lambda x: x,
            ScorerOutputType.PROBABILITY: lambda x: x.log(),
        }[context.state.model.outputType]

    def compute(self, scores, relevances, context):
        return -torch.logsumexp(
            self.normalize(scores) + (1 - 1.0 / relevances), -1
        ).sum() / len(scores)


class BatchwiseTrainer(LossTrainer):
    """Batchwise trainer

    Arguments:

    lossfn: The loss function to use
    sampler: A batchwise sampler
    """

    sampler: Param[BatchwiseSampler]
    """A batch-wise sampler"""

    lossfn: Param[BatchwiseLoss]
    """A batchwise loss function"""

    @initializer
    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)
        self.lossfn.initialize(context)
        self.sampler_iter = self.sampler.batchwise_iter(self.batch_size)

    def iter_batches(self) -> Iterator[BatchwiseRecords]:
        return self.sampler_iter

    def train_batch(self, batch: BatchwiseRecords):
        # Get the next batch and compute the scores for each query/document
        # Get the scores
        rel_scores = self.ranker(batch, self.context)

        if torch.isnan(rel_scores).any() or torch.isinf(rel_scores).any():
            self.logger.error("nan or inf relevance score detected. Aborting.")
            sys.exit(1)

        # Reshape to get the pairs and compute the loss
        batch_scores = rel_scores.reshape(*batch.relevances.shape)
        self.lossfn.process(
            batch_scores, batch.relevances.to(batch_scores.device), self.context
        )



from abc import ABC, abstractmethod
from typing import Iterator, Optional, List, Any
from experimaestro import Config, Param
import numpy as np
import datamaestro_text.data.ir.base as ir
from datamaestro_text.data.ir import DocumentStore, IDItem
from xpmir.datasets.adapters import TextStore
from xpmir.letor.samplers import PairwiseSampler
from xpmir.letor.records import (
    PairwiseRecords,
    PairwiseRecord,
)
from xpmir.utils.iter import (
    SerializableIterator,
    SkippingIterator,
    SerializableIteratorTransform,
)


class SampleTransform(Config, ABC):
    @abstractmethod
    def transform_topics(
        self, topics: List[ir.TopicRecord]
    ) -> Optional[List[ir.TopicRecord]]:
        ...

    @abstractmethod
    def transform_documents(
        self, documents: List[ir.DocumentRecord]
    ) -> Optional[List[ir.DocumentRecord]]:
        ...


class SampleHydrator(SampleTransform):
    """Base class for document/topic hydrators"""

    documentstore: Param[Optional[DocumentStore]]
    """The store for document texts if needed"""

    querystore: Param[Optional[TextStore]]
    """The store for query texts if needed"""

    def transform_topics(self, topics: List[ir.TopicRecord]):
        if self.querystore is None:
            return None
        return [
            ir.create_record(
                id=topic[IDItem].id, text=self.querystore[topic[IDItem].id]
            )
            for topic in topics
        ]

    def transform_documents(self, documents: List[ir.DocumentRecord]):
        if self.documentstore is None:
            return None
        return self.documentstore.documents_ext([d[IDItem].id for d in documents])


class SamplePrefixAdding(SampleTransform):
    """Transform the query and documents by adding the prefix"""

    query_prefix: Param[str] = ""
    """The prefix for the query"""

    document_prefix: Param[str] = ""
    """The prefix for the document"""

    def transform_topics(
        self, topics: List[ir.TopicRecord]
    ) -> Optional[List[ir.TopicRecord]]:
        if self.query_prefix == "" or len(topics) == 0:
            return None

        if isinstance(topics[0], ir.GenericTopic):
            return [
                ir.GenericTopic(topic[IDItem].id, self.query_prefix + topic.text)
                for topic in topics
            ]
        elif isinstance(topics[0], ir.TextTopic):
            return [ir.TextTopic(self.query_prefix + topic.text) for topic in topics]

    def transform_documents(
        self, documents: List[ir.DocumentRecord]
    ) -> Optional[List[ir.DocumentRecord]]:
        if self.document_prefix == "" or len(documents) == 0:
            return None

        if isinstance(documents[0], ir.GenericDocument):
            return [
                ir.GenericDocument(
                    document[IDItem].id, self.document_prefix + document.text
                )
                for document in documents
            ]
        elif isinstance(documents[0], ir.TextDocument):
            return [
                ir.TextDocument(self.document_prefix + document.text)
                for document in documents
            ]


class SampleTransformList(SampleTransform):
    """A class which group a list of sample transforms"""

    adapters: Param[List[SampleTransform]]
    """The list of sample transform to be applied"""

    def transform_topics(self, topics: List[ir.TopicRecord]) -> List[ir.TopicRecord]:
        for adapter in self.adapters:
            topics = adapter.transform_topics(topics) or topics
        return topics

    def transform_documents(
        self, documents: List[ir.DocumentRecord]
    ) -> List[ir.DocumentRecord]:
        for adapter in self.adapters:
            documents = adapter.transform_documents(documents) or documents
        return documents


class PairwiseTransformAdapter(PairwiseSampler):
    """Transforms pairwise samples using an adapter

    It is interesting to use this adapter since the transformation is only
    performed if the samples are used: when using a SkippingIterator, when
    recovering a checkpoint, all the records might have to be processed
    otherwise.
    """

    sampler: Param[PairwiseSampler]
    """The distillation samples without texts for query and documents"""

    adapter: Param[SampleTransform]
    """The transformation"""

    def initialize(self, random: Optional[np.random.RandomState] = None):
        super().initialize(random)
        self.sampler.initialize(random)

    def transform_record(self, record: PairwiseRecord) -> PairwiseRecord:
        topics = [record.query]
        docs = [record.positive, record.negative]

        topics = self.adapter.transform_topics(topics) or topics
        docs = self.adapter.transform_documents(docs) or docs

        return PairwiseRecord(topics[0], docs[0], docs[1])

    def pairwise_iter(self) -> Iterator[PairwiseRecord]:
        iterator = self.sampler.pairwise_iter()

        return SerializableIteratorTransform(
            SkippingIterator.make_serializable(iterator), self.transform_record
        )

    def transform_records(self, records: PairwiseRecords) -> PairwiseRecords:
        if topics := self.adapter.transform_topics(
            topic for topic in records.unique_topics
        ):
            records.set_unique_topics(topics)

        if documents := self.adapter.transform_documents(
            document for document in records.unique_documents
        ):
            records.set_unique_documents(documents)
        return records

    def pairwise_batch_iter(self, size) -> SerializableIterator[PairwiseRecords, Any]:
        iterator = self.sampler.pairwise_batch_iter(size)
        return SerializableIteratorTransform(
            SkippingIterator.make_serializable(iterator), self.transform_records
        )



import torch
import json
from typing import Dict, Any
from pathlib import Path

from experimaestro import Annotated, Param, pathgenerator, field

from xpmir.learning import SampleIterator
from xpmir.learning.batchers import Batcher
from xpmir.learning.learner import (
    Learner,
    LearnerListener,
    LearnerListenerStatus,
    TrainerContext,
    TrainState,
)
from xpmir.letor.learner import ValidationModuleLoader
from xpmir.letor.records import BaseRecords
from xpmir.letor.trainers import LossTrainer
from xpmir.learning.metrics import Metrics, ScalarMetric
from xpmir.utils.logging import easylog

logger = easylog()


class TrainerValidationLoss(LearnerListener):
    """Generic trainer-based loss validation"""

    data: Param[SampleIterator]
    """The dataset to use"""

    batcher: Param[Batcher] = field(default_factory=Batcher.C)
    """How to batch samples together"""

    batch_size: Param[int]
    """Batch size"""

    trainer: Param[LossTrainer]
    """The trainer"""

    warmup: Param[int] = -1
    """How many epochs before actually computing the validation loss"""

    bestpath: Annotated[Path, pathgenerator("best")]
    """Path to the best checkpoints"""

    info: Annotated[Path, pathgenerator("info.json")]
    """Path to the JSON file that contains the metric values at each epoch"""

    validation_interval: Param[int] = 1
    """Epochs between each validation"""

    early_stop: Param[int] = 0
    """Number of epochs without improvement after which we stop learning.
    Should be a multiple of validation_interval or 0 (no early stopping)"""

    def __validate__(self):
        assert (
            self.early_stop % self.validation_interval == 0
        ), "Early stop should be a multiple of the validation interval"

    def initialize(self, learner: Learner, context: TrainerContext):
        super().initialize(learner, context)
        self.scope = f"validation/{self.id}"
        self.bestpath.mkdir(exist_ok=True, parents=True)
        self.batcher_worker = self.batcher.initialize(self.batch_size)

        # Checkpoint start
        try:
            with self.info.open("rt") as fp:
                self.top: Dict[str, Any] = json.load(fp)
        except Exception:
            self.top = None

    def init_task(self, learner: "Learner", dep):
        return dep(
            ValidationModuleLoader.C(
                value=learner.model,
                listener=self,
                key="",
                path=self.bestpath / TrainState.MODEL_PATH,
            )
        )

    def update_metrics(self, metrics: Dict[str, float]):
        if self.top:
            # Just use another key
            metrics[f"{self.id}/final"] = self.top["value"]

    def should_stop(self, epoch=0):
        if self.early_stop > 0 and self.top:
            epochs_since_imp = (epoch or self.context.epoch) - self.top["epoch"]
            if epochs_since_imp >= self.early_stop:
                return LearnerListenerStatus.STOP

        return LearnerListenerStatus.DONT_STOP

    def reducer(self, records: BaseRecords, metrics: Metrics):
        """Combines a forward and backard

        This method can be implemented by specific trainers that use the gradient.
        In that case the regularizer losses should be taken into account with
        `self.add_losses`.
        """
        # Restrict losses to this context

        with self.context.losses() as losses:
            # Compute the loss(es)
            self.trainer.train_batch(records)

            # Aggregate with previous values
            nrecords = len(records)
            total_loss = 0.0
            names = []

            for loss in losses:
                total_loss += loss.weight * loss.value
                names.append(loss.name)
                metrics.add(
                    ScalarMetric(
                        f"loss/{loss.name}", float(loss.value.item()), nrecords
                    )
                )

            # Reports the main metric
            if len(names) > 1:
                names.sort()
                metrics.add(ScalarMetric("loss", float(total_loss.item()), nrecords))

            return metrics

    def __call__(self, state: TrainState):
        # Check that we did not stop earlier (when loading from checkpoint / if other
        # listeners have not stopped yet)
        if self.should_stop(state.epoch - 1) == LearnerListenerStatus.STOP:
            return LearnerListenerStatus.STOP

        if state.epoch % self.validation_interval == 0:
            # Compute validation metrics
            metrics = Metrics()
            state.model.eval()

            with torch.no_grad():
                for batch in self.data.__batch_iter__(self.batch_size):
                    self.batcher_worker.reduce(
                        batch, self.reducer, metrics, raise_oom=False
                    )

            metrics.report(state.step, self.context.writer, self.id)

            # Get the current value
            if len(metrics.metrics) == 1:
                value = next(iter(metrics.metrics.values())).compute()
            else:
                value = metrics.metrics["loss"].compute()

            # Update the top validation
            if state.epoch >= self.warmup:
                topstate = self.top
                if topstate is None or value > topstate["value"]:
                    # Save the new top JSON
                    self.top = {"value": value, "epoch": self.context.epoch}

                    # Copy in corresponding directory
                    logger.info(f"Saving the checkpoint {state.epoch}")
                    self.context.copy(self.bestpath)

            # Update information
            with self.info.open("wt") as fp:
                json.dump(self.top, fp)

        # Early stopping?
        return self.should_stop()



import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from experimaestro import Config, Param
from experimaestro.compat import cached_property
import torch
from experimaestro.taskglobals import Env as TaskEnv
import torch.distributed as dist
import torch.multiprocessing as mp
import tempfile
from xpmir.context import Context
from xpmir.utils.utils import easylog

logger = easylog()


@dataclass
class DeviceInformation:
    device: torch.device
    """The device"""

    main: bool
    """Flag for the main process (all other are slaves)"""

    count: int = 1
    """Number of processes"""

    rank: int = 0
    """When using distributed processing, this is the rank of the process"""


class ComputationContext(Context):
    device_information: DeviceInformation


@dataclass
class DistributedDeviceInformation(DeviceInformation):
    pass


class Device(Config):
    """Device to use, as well as specific option (e.g. parallelism)"""

    @cached_property
    def value(self):
        import torch

        return torch.device("cpu")

    n_processes = 1
    """Number of processes"""

    def execute(self, callback, *args, **kwargs):
        callback(DeviceInformation(self.value, True), *args, **kwargs)


def mp_launcher(rank, path, world_size, callback, taskenv, args, kwargs):
    logger.info("Started process for rank %d [%s]", rank, path)
    TaskEnv._instance = taskenv
    taskenv.slave = rank == 0

    logger.info("Initializing process group [%d]", rank)
    dist.init_process_group(
        "gloo", init_method=f"file://{path}", rank=rank, world_size=world_size
    )

    logger.info("Calling callback [%d]", rank)
    device = torch.device(f"cuda:{rank}")
    callback(
        DistributedDeviceInformation(
            device=device, main=rank == 0, rank=rank, count=world_size
        ),
        *args,
        **kwargs,
    )

    # Cleanup
    dist.destroy_process_group()


def cuda_execute(callback, args, kwargs, distributed=True):
    # Setup distributed computation
    # Seehttps://pytorch.org/tutorials/intermediate/ddp_tutorial.html
    n_gpus = torch.cuda.device_count()

    if n_gpus == 1 or not distributed:
        logger.info("Using mono-GPU CUDA computing")
        callback(DeviceInformation(torch.device("cuda"), True), *args, **kwargs)
    else:
        if sys.version_info.major == 3 and sys.version_info.minor < 10:
            tmp_directory = tempfile.TemporaryDirectory()
        else:
            tmp_directory = tempfile.TemporaryDirectory(ignore_cleanup_errors=True)

        with tmp_directory as directory:
            logger.info("Setting up distributed CUDA computing (%d GPUs)", n_gpus)
            return mp.start_processes(
                mp_launcher,
                args=(
                    str((Path(directory) / "link").absolute()),
                    n_gpus,
                    callback,
                    TaskEnv.instance(),
                    args,
                    kwargs,
                ),
                nprocs=n_gpus,
                join=True,
                start_method=mp.get_start_method(),
            )


class CudaDevice(Device):
    """CUDA device"""

    gpu_determ: Param[bool] = False
    """Sets the deterministic"""

    cpu_fallback: Param[bool] = False
    """Fallback to CPU if no GPU is available"""

    distributed: Param[bool] = False
    """Flag for using DistributedDataParallel

    When the number of GPUs is greater than one, use
    torch.nn.parallel.DistributedDataParallel when `distributed` is `True` and
    the number of GPUs greater than 1. When False, use `torch.nn.DataParallel`
    """

    @cached_property
    def value(self):
        """Called by experimaestro to substitute object at run time"""
        if not torch.cuda.is_available():
            if not self.cpu_fallback:
                # Not accepting fallbacks
                raise AssertionError("No GPU available")
            logger.error("No GPU available. Falling back on CPU.")
            return torch.device("cpu")

        # Set the deterministic flag
        torch.backends.cudnn.deterministic = self.gpu_determ
        if self.gpu_determ:
            logger.debug("using GPU (deterministic)")
        else:
            logger.debug("using GPU (non-deterministic)")

        return torch.device("cuda")

    @cached_property
    def n_processes(self):
        """Number of processes"""
        if self.distributed:
            return torch.cuda.device_count()
        return 1

    def execute(self, callback, *args, **kwargs):
        cuda_execute(callback, args, kwargs)


class BestDevice(Device):
    """Try to use a GPU device if it exists, fallbacks to CPU otherwise

    To be used when debugging"""

    @cached_property
    def value(self):
        import torch

        if torch.cuda.is_available():
            device = torch.device("cuda")
            logging.info(f"Using GPU: {torch.cuda.get_device_name(0)}")
        elif torch.backends.mps.is_available():
            device = torch.device("mps")
            logging.info(f"Using MPS: {torch.backends.mps.is_available()}")
        else:
            device = torch.device("cpu")
            logging.info(f"Using CPU: {torch.device('cpu')}")
        return device

    def execute(self, callback, *args, **kwargs):
        if self.value.type != "cuda":
            callback(DeviceInformation(self.value, True), *args, **kwargs)
        else:
            cuda_execute(callback, args, kwargs)

    @cached_property
    def n_processes(self):
        """Number of processes"""
        if torch.cuda.is_available():
            return torch.cuda.device_count()
        return 1


# Default device is the CPU
DEFAULT_DEVICE = Device.C()



import torch


def lengthToMask(length: torch.LongTensor, max_len=None, dtype=None):
    """length: B.
    return B x max_len.
    If max_len is None, then max of length will be used.
    """
    assert len(length.shape) == 1, "Length shape should be 1 dimensional."
    max_len = max_len or length.max().item()
    mask = torch.arange(max_len, device=length.device, dtype=length.dtype).expand(
        len(length), max_len
    ) < length.unsqueeze(1)
    if dtype is not None:
        mask = torch.as_tensor(mask, dtype=dtype, device=length.device)
    return mask



from typing import List, Tuple, Iterable
from experimaestro import Param
from xpmir.rankers import Scorer, ScoredDocument
from xpmir.letor.samplers.hydrators import SampleTransform
from xpmir.letor.records import TopicRecord
from xpmir.learning import ModuleInitOptions


class ScorerTransformAdapter(Scorer):
    """Transforms topic and/or documents output by a scorer when rescoring documents"""

    scorer: Param[Scorer]
    """The original scorer to be transform"""

    adapter: Param[SampleTransform]
    """The list of sample transforms to apply"""

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)
        self.scorer.__initialize__(options)

    def eval(self):
        self.scorer.train(False)

    def to(self, device):
        self.scorer.to(device)

    def transform_records(
        self, topic: TopicRecord, scored_documents: Iterable[ScoredDocument]
    ) -> Tuple[TopicRecord, List[ScoredDocument]]:
        topics = [topic.topic]
        docs = [sd.document for sd in scored_documents]

        topics = self.adapter.transform_topics(topics) or topics
        docs = self.adapter.transform_documents(docs) or docs

        sd_list = [
            ScoredDocument(doc, sd.score) for (doc, sd) in zip(docs, scored_documents)
        ]

        return (topics[0].as_record(), sd_list)

    def compute(
        self, topic: TopicRecord, documents: Iterable[ScoredDocument]
    ) -> List[ScoredDocument]:
        topic, sd_list = self.transform_records(topic, documents)
        return self.scorer.compute(topic, sd_list)



from typing import Any
from dataclasses import InitVar
import sys
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from experimaestro import Param, Config
from xpmir.letor.records import PointwiseRecords, PointwiseRecord
from xpmir.letor.trainers import LossTrainer
from xpmir.learning.context import Loss, TrainerContext
from xpmir.rankers import LearnableScorer, ScorerOutputType
from xpmir.letor.samplers import PointwiseSampler, SerializableIterator


class PointwiseLoss(Config):
    NAME = "?"
    weight: Param[float] = 1.0

    def initialize(self, scorer: LearnableScorer):
        pass

    def process(self, scores, targets, context: TrainerContext):
        value = self.compute(scores, targets)
        context.add_loss(Loss(f"point-{self.NAME}", value, self.weight))

    def compute(self, rel_scores, target_relscores) -> torch.Tensor:
        raise NotImplementedError()


class MSELoss(PointwiseLoss):
    NAME = "mse"

    def compute(self, rel_scores, target_relscores):
        return F.mse_loss(rel_scores, target_relscores)


class BinaryCrossEntropyLoss(PointwiseLoss):
    """Computes binary cross-entropy

    Uses a BCE with logits if the scorer output type is
    not a probability
    """

    NAME = "bce"

    def initialize(self, scorer: LearnableScorer):
        if scorer.outputType == ScorerOutputType.PROBABILITY:
            self.loss = nn.BCELoss()
        else:
            self.loss = nn.BCEWithLogitsLoss()

    def compute(self, rel_scores, target_relscores):
        device = target_relscores.device
        return self.loss(rel_scores.to(device), (target_relscores > 0).float())


class PointwiseTrainer(LossTrainer):
    """Pointwise trainer"""

    lossfn: Param[PointwiseLoss] = MSELoss()
    """Loss function to use"""

    sampler: Param[PointwiseSampler]
    """The pairwise sampler"""

    sampler_iter: InitVar[SerializableIterator[PointwiseRecord, Any]]

    def initialize(self, random: np.random.RandomState, context):
        super().initialize(random, context)

        self.sampler.initialize(self.random)
        self.lossfn.initialize(self.ranker)

        self.random = random
        self.sampler_iter = self.sampler.pointwise_iter()

    def __validate__(self):
        # assert self.grad_acc_batch >= 0, "Adaptative batch size not implemented"
        pass

    def iter_batches(self):
        while True:
            batch = PointwiseRecords()
            for _, record in zip(range(self.batch_size), self.sampler_iter):
                batch.add(record)

            yield batch

    def train_batch(self, records: PointwiseRecords):

        rel_scores = self.ranker(records, self.context)

        if torch.isnan(rel_scores).any() or torch.isinf(rel_scores).any():
            self.logger.error("nan or inf relevance score detected. Aborting.")
            sys.exit(1)

        target_relscores = torch.FloatTensor(records.relevances)

        rel_scores = rel_scores.flatten()

        self.lossfn.process(rel_scores, target_relscores, self.context)

        # TODO: Create classes for the missing losses
        # Apply the loss
        # elif self.lossfn == "mse-nil":
        #     loss = F.mse_loss(
        #         rel_scores.flatten(), torch.zeros_like(rel_scores.flatten())
        #     )
        # elif self.lossfn == "l1":
        #     loss = F.l1_loss(rel_scores.flatten(), target_relscores)
        # elif self.lossfn == "l1pos":
        #     loss = F.l1_loss(rel_scores.flatten(), (target_relscores > 0.0).float())
        # elif self.lossfn == "smoothl1":
        #     loss = F.smooth_l1_loss(rel_scores.flatten(), target_relscores)
        # elif self.lossfn == "cross_entropy":
        #     loss = -torch.where(
        #         target_relscores > 0, rel_scores.flatten(), 1 - rel_scores.flatten()
        #     ).log()
        #     loss = loss.mean()
        # elif self.lossfn == "cross_entropy_logits":
        #     assert len(rel_scores.shape) == 2
        #     assert rel_scores.shape[1] == 2
        #     log_probs = -rel_scores.log_softmax(dim=1)
        #     one_hot = torch.tensor(
        #         [[1.0, 0.0] if tar > 0 else [0.0, 1.0] for tar in target_relscores],
        #         device=rel_scores.device,
        #     )
        #     loss = (log_probs * one_hot).sum(dim=1).mean()
        # elif self.lossfn == "softmax":
        #     assert len(rel_scores.shape) == 2
        #     assert rel_scores.shape[1] == 2
        #     probs = rel_scores.softmax(dim=1)
        #     one_hot = torch.tensor(
        #         [[0.0, 1.0] if tar > 0 else [1.0, 0.0] for tar in target_relscores],
        #         device=rel_scores.device,
        #     )
        #     loss = (probs * one_hot).sum(dim=1).mean()
        # elif self.lossfn == "mean":
        #     loss = rel_scores.mean()
        # else:
        #     raise ValueError(f"unknown lossfn `{self.lossfn}`")



import os
from pathlib import Path
import logging
from typing import List, Optional, Tuple, Union
from functools import cached_property
import torch
from experimaestro import Config, Param

from xpmir.learning.optim import ModuleInitOptions
from xpmir.utils.utils import Initializable
from xpmir.utils.convert import Converter
from xpmir.text.tokenizers import (
    TokenizerBase,
    TokenizedTexts,
    TokenizerInput,
    TokenizerOptions,
)

try:
    from transformers import AutoTokenizer
except Exception:
    logging.error("Install huggingface transformers to use these configurations")
    raise


HFTokenizerInput = Union[List[str], List[Tuple[str, str]]]


class HFTokenizer(Config, Initializable):
    """This is the main tokenizer class"""

    model_id: Param[str]
    """The tokenizer hugginface ID"""

    max_length: Param[int] = 4096
    """Maximum length for the tokenizer (can be overridden by the model)"""

    DEFAULT_OPTIONS = TokenizerOptions()

    def __initialize__(self, options: ModuleInitOptions):
        """Initialize the HuggingFace transformer"""

        model_id_or_path = self.model_id

        # Use saved models
        if model_path := os.environ.get("XPMIR_TRANSFORMERS_CACHE", None):
            path = Path(model_path) / "tokenizers" / Path(self.model_id)
            if path.is_dir():
                logging.warning("Using saved tokenizer from %s", path)
                model_id_or_path = path
            else:
                logging.warning(
                    "Could not find saved tokenizer in %s, using HF loading", path
                )

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_id_or_path, model_max_length=self.max_length
        )

        self.cls = self.tokenizer.cls_token
        self.cls_id = self.tokenizer.cls_token_id
        self.sep_id = self.tokenizer.sep_token_id

    def tokenize(
        self,
        texts: HFTokenizerInput,
        options: Optional[TokenizerOptions] = None,
    ) -> TokenizedTexts:
        options = options or HFTokenizer.DEFAULT_OPTIONS
        max_length = options.max_length
        if max_length is None:
            max_length = self.tokenizer.model_max_length
        else:
            max_length = min(max_length, self.maxtokens())

        r = self.tokenizer(
            list(texts),
            max_length=max_length,
            truncation=True,
            padding=True,
            return_tensors="pt",
            return_length=options.return_length,
            return_attention_mask=options.return_mask,
        )

        return TokenizedTexts(
            None,
            r["input_ids"],
            r["length"],
            r.get("attention_mask", None),
            r.get("token_type_ids", None),
        )

    def id2tok(self, idx):
        """Returns the token strings corresponding to the token ids"""
        if torch.is_tensor(idx):
            if len(idx.shape) == 0:
                return self.id2tok(idx.item())
            return [self.id2tok(x) for x in idx]
        return self.tokenizer.id_to_token(idx)

    def lexicon_size(self) -> int:
        return self.tokenizer._tokenizer.get_vocab_size()

    def maxtokens(self) -> int:
        return self.tokenizer.model_max_length

    def dim(self):
        return self.tokenizer.config.hidden_size

    @property
    def vocab_size(self) -> int:
        """Returns the size of the vocabulary"""
        return self.tokenizer.vocab_size

    @property
    def vocab(self) -> dict:
        return self.tokenizer.vocab


class HFTokenizerBase(TokenizerBase[TokenizerInput, TokenizedTexts]):
    """Base class for all Hugging-Face tokenizers"""

    tokenizer: Param[HFTokenizer]
    """The HuggingFace tokenizer"""

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)
        self.tokenizer.initialize(options)

    @classmethod
    def from_pretrained_id(cls, hf_id: str, **kwargs):
        return cls(tokenizer=HFTokenizer(model_id=hf_id), **kwargs)

    def vocabulary_size(self):
        return self.tokenizer.vocab_size

    def tok2id(self, tok: str) -> int:
        return self.tokenizer.tok2id(tok)

    def id2tok(self, idx: int) -> str:
        return self.tokenizer.id2tok(idx)
    
    def get_vocabulary(self):
        return self.tokenizer.vocab


class HFStringTokenizer(HFTokenizerBase[HFTokenizerInput]):
    """Process list of texts"""

    def tokenize(
        self, texts: List[HFTokenizerInput], options: Optional[TokenizerOptions] = None
    ) -> TokenizedTexts:
        return self.tokenizer.tokenize(texts, options=options)


class HFTokenizerAdapter(HFTokenizerBase[TokenizerInput]):
    """Process list of texts"""

    converter: Param[Converter[TokenizerInput, HFTokenizerInput]]

    def tokenize(
        self, inputs: List[TokenizerInput], options: Optional[TokenizerOptions] = None
    ) -> TokenizedTexts:
        return self.tokenizer.tokenize(
            [self.converter(input) for input in inputs], options=options
        )


class HFListTokenizer(HFTokenizerBase[List[List[str]]]):
    """Process list of texts by separating them by a separator token"""

    separate_index: Param[bool] = 0
    """Use a tuple until this index"""

    @cached_property
    def sep_string(self):
        token = self.tokenizer.tokenizer.sep_token
        assert token is not None
        return token

    def join(self, text_list: List[str]):
        if self.separate_index > 0:
            r = text_list[: self.separate_index]
            r.extend([self.sep_string.join(text_list[self.separate_index :])])
            return r

        return self.sep_string.join(text_list)

    def tokenize(
        self, text_lists: List[List[str]], options: Optional[TokenizerOptions] = None
    ) -> TokenizedTexts:
        return self.tokenizer.tokenize(
            [self.join(text_list) for text_list in text_lists],
            options=options,
        )



from dataclasses import InitVar
import sys
from typing import Iterator

import numpy as np
import torch
import torch.nn.functional as F
from xpmir.learning.context import Loss, TrainerContext
from xpmir.letor.records import DocumentRecords, DocumentRecord
from xpmir.letor.trainers import LossTrainer

from experimaestro import Param, Config

from xpmir.letor.samplers import Sampler
from xpmir.utils.iter import RandomSerializableIterator


class MLMLoss(Config):
    NAME = "?"
    weight: Param[float] = 1.0

    def initialize(self):
        pass

    def process(self, scores, targets, context: TrainerContext):
        value = self.compute(scores, targets)
        context.add_loss(Loss(f"point-{self.NAME}", value, self.weight))

    def compute(self, rel_scores, target_relscores) -> torch.Tensor:
        raise NotImplementedError()


class CrossEntropyLoss(MLMLoss):
    """Computes cross-entropy

    Uses a CE with logits if the scorer output type is
    not a probability
    """

    NAME = "ce"

    def compute(self, rel_scores, target_relscores):
        return F.cross_entropy(rel_scores, target_relscores)


class MLMTrainer(LossTrainer):
    """Trainer for Masked Language Modeling"""

    # Loss function to use
    lossfn: Param[MLMLoss] = CrossEntropyLoss()

    sampler: Param[Sampler]

    sampler_iter: InitVar[RandomSerializableIterator[DocumentRecord]]

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)
        self.lossfn.initialize()
        self.sampler.initialize(random)
        self.sampler_iter = self.sampler.record_iter()

    def __validate__(self):
        # assert self.grad_acc_batch >= 0, "Adaptative batch size not implemented"
        pass

    def iter_batches(self) -> Iterator[DocumentRecords]:
        while True:
            batch = DocumentRecords()
            for _, record in zip(range(self.batch_size), self.sampler_iter):
                batch.add(record)

            yield batch

    def train_batch(self, records: DocumentRecords):
        mlm_output = self.model(records.to_texts(), self.context)

        if torch.isnan(mlm_output.logits).any() or torch.isinf(mlm_output.logits).any():
            self.logger.error("nan or inf relevance score detected. Aborting.")
            sys.exit(1)

        self.lossfn.process(
            mlm_output.logits.view(-1, self.model.vocab_size),
            mlm_output.labels.view(-1).to(mlm_output.logits.device),
            self.context,
        )



from abc import ABC, abstractmethod
from typing import Generic, TypeVar
from experimaestro import Config

Input = TypeVar("Input")
Output = TypeVar("Output")


class Converter(Config, ABC, Generic[Input, Output]):
    @abstractmethod
    def __call__(self, input: Input) -> Output:
        pass



# See documentation on https://datamaestro.readthedocs.io

from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro_text.data.ir.huggingface import HuggingFacePairwiseSampleDataset
from datamaestro.download.huggingface import hf_download


@datatags("information retrieval", "hard negatives", "msmarco")
@datatasks("learning to rank")
@hf_download(
    "dataset",
    "sentence-transformers/msmarco-hard-negatives",
    data_files="msmarco-hard-negatives-msmarco.jsonl.gz",
    split="train",
)
@dataset(
    HuggingFacePairwiseSampleDataset,
    url="https://huggingface.co/datasets/sentence-transformers/msmarco-hard-negatives",
)
def ensemble(dataset):
    """Hard negatives mined from a set of models"""
    return dataset



"""Interface to the Facebook FAISS library

https://github.com/facebookresearch/faiss
"""

from pathlib import Path
from typing import Callable, Iterator, List, Optional, Tuple
from datamaestro_text.data.ir.base import TopicRecord
from experimaestro import Config, initializer
import torch
import numpy as np
from experimaestro import Annotated, Meta, Task, pathgenerator, Param, tqdm
import logging
from datamaestro_text.data.ir import DocumentStore, TextItem
from xpmir.rankers import Retriever, ScoredDocument
from xpmir.learning.batchers import Batcher
from xpmir.learning import ModuleInitMode
from xpmir.text.encoders import TextEncoder
from xpmir.letor import (
    Device,
    DEFAULT_DEVICE,
    DeviceInformation,
)
from xpmir.utils.utils import batchiter, easylog, foreach
from xpmir.documents.samplers import DocumentSampler
from xpmir.context import Context, Hook, InitializationHook

logger = easylog()

try:
    import faiss
except ModuleNotFoundError:
    logging.error("FAISS library is not available (install faiss-cpu or faiss)")
    raise


class FaissIndex(Config):
    """FAISS Index"""

    normalize: Param[bool]
    """Whether vectors should be normalized (L2)"""

    faiss_index: Annotated[Path, pathgenerator("faiss.dat")]
    """Path to the file containing the index"""

    documents: Param[DocumentStore]
    """The set of documents"""


class IndexBackedFaiss(FaissIndex, Task):
    """Constructs a FAISS index backed up by an index

    During executions, InitializationHooks are used (pre/post)

    """

    encoder: Param[TextEncoder]
    """Encoder for document texts"""

    batchsize: Meta[int] = 1
    """The batch size used when computing representations of documents"""

    device: Meta[Device] = DEFAULT_DEVICE
    """The device used by the encoder"""

    batcher: Meta[Batcher] = Batcher()
    """The way to prepare batches of documents"""

    hooks: Param[List[Hook]] = []
    """An optional list of hooks"""

    indexspec: Param[str]
    """The index type as a factory string

    See https://github.com/facebookresearch/faiss/wiki/Faiss-indexes
    for the full list of indices

    and https://github.com/facebookresearch/faiss/wiki/The-index-factory
    for the combination of the index factory
    """

    sampler: Param[Optional[DocumentSampler]]
    """Optional document sampler when training the index -- by default, all the
    documents from the collection are used"""

    def full_sampler(self) -> Tuple[int, Iterator[str]]:
        """Returns an iterator over the full set of documents"""
        iter = (d for d in self.documents.iter_documents())
        return self.documents.documentcount or 0, iter

    def train(
        self,
        index: faiss.Index,
        batch_encoder: Callable[[Iterator[str]], Iterator[torch.Tensor]],
    ):
        """train the index

        params
        index:
        batch_encoder:
            function, input is a iterator of list of documents str, return the
            encoded document vector(tensor of shape (bs*dimension))
        """
        logger.info("Building index")
        count, iter = (
            self.sampler() if self.sampler is not None else self.full_sampler()
        )
        doc_iter = tqdm(
            iter, total=count, desc="Collecting the representation of documents (train)"
        )

        # Collect batches (in memory)
        logger.info("Collecting the representation of %d documents", count)
        sample = np.ndarray((count, self.encoder.dimension), dtype=np.float32)
        ix = 0
        for batch in batch_encoder(doc_iter):
            sample[ix : (ix + len(batch))] = batch.cpu().numpy()
            ix += len(batch)

        logger.info("Training index (%d samples)", count)
        # Here we may use just a part of the document to train the index
        index.train(sample)

    def execute(self):
        self.device.execute(self._execute)

    def _execute(self, device_information: DeviceInformation):
        # Initialization hooks
        context = Context(hooks=self.hooks)
        foreach(context.hooks(InitializationHook), lambda hook: hook.before(context))

        step_iter = tqdm(total=2, desc="Building the FAISS index")

        # Initializations
        self.encoder.initialize(ModuleInitMode.DEFAULT.to_options())
        index = faiss.index_factory(
            self.encoder.dimension, self.indexspec, faiss.METRIC_INNER_PRODUCT
        )
        batcher = self.batcher.initialize(self.batchsize)

        # Change the device of the encoder
        self.encoder.to(device_information.device).eval()

        # Train the index
        if not index.is_trained:
            with torch.no_grad():
                logging.info("Training FAISS index (%d documents)", index.ntotal)

                def batch_encoder(doc_iter: Iterator[str]):
                    for batch in batchiter(self.batchsize, doc_iter):
                        data = []
                        batcher.process(batch, self.encode, data)
                        yield torch.cat(data)

                self.train(index, batch_encoder)

        step_iter.update()

        # Index the collection
        doc_iter = (
            tqdm(
                self.documents.iter_documents(),
                total=self.documents.documentcount,
                desc="Indexing the collection",
            )
            if device_information.main
            else self.documents.iter_documents()
        )

        # Initialization hooks (after)
        foreach(context.hooks(InitializationHook), lambda hook: hook.after(context))

        # Let's index !
        # We add index for all the documents
        with torch.no_grad():
            for batch in batchiter(self.batchsize, doc_iter):
                batcher.process(
                    [document[TextItem].text for document in batch],
                    self.index_documents,
                    index,
                )

        logging.info("Writing FAISS index (%d documents)", index.ntotal)
        faiss.write_index(index, str(self.faiss_index))
        step_iter.update()

    def encode(self, batch: List[str], data: List):
        batch = [
            text for text in batch if text != ""
        ]  # remove the empty strings in the dataset (training only)
        x = self.encoder(batch)
        if self.normalize:
            x /= x.norm(2, keepdim=True, dim=1)
        data.append(x)

    def index_documents(self, batch: List[str], index):
        x = self.encoder(batch).value
        if self.normalize:
            x /= x.norm(2, keepdim=True, dim=1)
        index.add(np.ascontiguousarray(x.cpu().numpy()))


class FaissRetriever(Retriever):
    """Retriever based on Faiss"""

    encoder: Param[TextEncoder]
    """The query encoder"""

    index: Param[FaissIndex]
    """The faiss index"""

    topk: Param[int]
    """the number of documents to be retrieved"""

    @initializer
    def initialize(self):
        logger.info("FAISS retriever (1/2): initializing the encoder")
        self.encoder.initialize()
        logger.info("FAISS retriever (2/2): reading the index")
        self._index = faiss.read_index(str(self.index.faiss_index))
        logger.info("FAISS retriever: initialized")

    def retrieve(self, query: TopicRecord) -> List[ScoredDocument]:
        """Retrieves a documents, returning a list sorted by decreasing score"""
        with torch.no_grad():
            self.encoder.eval()  # pass the model to the evaluation model
            encoded_query = self.encoder([query[TextItem].text]).value
            if self.index.normalize:
                encoded_query /= encoded_query.norm(2)

            values, indices = self._index.search(encoded_query.cpu().numpy(), self.topk)
            return [
                ScoredDocument(self.index.documents.document_int(int(ix)), float(value))
                for ix, value in zip(indices[0], values[0])
                if ix >= 0
            ]



from typing import List, Optional
from experimaestro import Constant, Param
from torch import nn
from xpmir.text import (
    TokensEncoderOutput,
    TokenizedTextEncoderBase,
    TokenizerOptions,
)
from xpmir.text.encoders import InputType
from xpmir.neural.interaction import (
    InteractionScorer,
    SimilarityInput,
    SimilarityOutput,
    TrainerContext,
)
from xpmir.neural.interaction.common import CosineSimilarity


class Colbert(InteractionScorer):
    """ColBERT model

    Implementation of the Colbert model from:

        Khattab, Omar, and Matei Zaharia. “ColBERT: Efficient and Effective
        Passage Search via Contextualized Late Interaction over BERT.” SIGIR
        2020, Xi'An, China

    For the standard Colbert model, use the colbert function
    """

    version: Constant[int] = 2
    """Current version of the code (changes if a bug is found)"""

    linear_dim: Param[int] = 128
    """Size of the last linear layer (before computing inner products)"""

    compression_size: Param[int] = 128
    """Projection layer for the last layer (or 0 if None)"""

    def __validate__(self):
        super().__validate__()

        assert self.compression_size >= 0, "Last layer size should be 0 or above"

    def __initialize__(self, options):
        super().__initialize__(options)
        self.linear = nn.Linear(self.encoder.dimension, self.linear_dim, bias=False)

    def _encode(
        self,
        texts: List[InputType],
        encoder: TokenizedTextEncoderBase[InputType, TokensEncoderOutput],
        options: TokenizerOptions,
    ) -> SimilarityInput:
        encoded = encoder(texts, options=options)
        return self.similarity.preprocess(
            SimilarityInput(self.linear(encoded.value), encoded.tokenized.mask)
        )

    def compute_scores(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
        value: SimilarityOutput,
        info: Optional[TrainerContext] = None,
    ):
        # Similarity matrix B x Lq x Ld or Bq x Lq x Bd x Ld
        s = value.similarity.masked_fill(
            value.d_view(documents.mask).logical_not(), float("-inf")
        ).masked_fill(value.q_view(queries.mask).logical_not(), 0)
        return s.max(-1).values.sum(1).flatten()


def colbert(
    model_id: str,
    *,
    query_token: bool = False,
    doc_token: bool = False,
    mask_token: bool = True,
):
    """Creates standard ColBERT model based on a HuggingFace transformer

    :param model_id: The HF model ID
    :param query_token: Whether to use a query prefix token when encoding
        queries, defaults to False
    :param doc_token: Whether to use a document prefix token to encode
        documents, defaults to False
    :param mask_token: Whether to use a mask tokens to encode queries, defaults
        to True
    :return: A ColBERT configuration object
    """
    from xpmir.text.encoders import TokenizedTextEncoder
    from xpmir.text.adapters import TopicTextConverter
    from xpmir.text.huggingface import HFTokensEncoder
    from xpmir.text.huggingface.tokenizers import HFTokenizerAdapter

    assert query_token is False, "Not implemented: use [QUERY] token"
    assert doc_token is False, "Not implemented: use [DOCUMENT] token"
    assert mask_token is False, "Not implemented: use [MASK] token"

    encoder = TokenizedTextEncoder(
        tokenizer=HFTokenizerAdapter.from_pretrained_id(
            model_id,
            converter=TopicTextConverter(),
        ),
        encoder=HFTokensEncoder.from_pretrained_id(model_id),
    )
    return Colbert(encoder=encoder, similarity=CosineSimilarity())



# Utility functions for MS-Marco experiments

from attrs import Factory
import logging

from . import NeuralIRExperiment, configuration
from xpmir.papers.helpers.samplers import ValidationSample

logging.basicConfig(level=logging.INFO)


@configuration()
class RerankerMSMarcoV1Configuration(NeuralIRExperiment):
    """Configuration for rerankers"""

    validation: ValidationSample = Factory(ValidationSample)


@configuration()
class DualMSMarcoV1Configuration(NeuralIRExperiment):
    """Configuration for sparse retriever"""

    validation: ValidationSample = Factory(ValidationSample)


@configuration()
class MLMMSMarcoV1Configuration(NeuralIRExperiment):
    """Configuration for sparse retriever"""

    validation: ValidationSample = Factory(ValidationSample)
